#!/usr/bin/env python3

import sys
import json
import glob
import os
import random
import re
from jsonschema import Draft202012Validator, validate

combined_base = os.environ.get("JSB_DATA") + "/unique_tests"


class Stats:
    def __init__(self):
        self.files = 0
        self.responses = 0
        self.server_error = 0
        self.schema_error = 0
        self.json_error = 0
        self.json_error_non_length = 0
        self.validation_error = 0
        self.invalidation_error = 0
        self.not_negative = 0
        self.negative_added = 0


stats = Stats()


def process_file(file_name):
    file_base = file_name.split("/")[-1]
    split = file_name.split("/")[-2]

    stats.files += 1

    with open(file_name) as f:
        schema = json.loads(f.read())

    combined_name = f"{combined_base}/{split}---{file_base}"
    combined = {
        "description": f"sample {split}/{file_base}",
        "schema": schema,
        "tests": [],
    }

    if os.path.exists(combined_name):
        with open(combined_name, "r") as f:
            combined = json.loads(f.read())
            prev = json.dumps(combined["schema"])
            combined["schema"] = schema
            if json.dumps(combined["schema"]) == prev:
                return  # no change

    try:
        Draft202012Validator.check_schema(schema)
    except Exception as e:
        stats.schema_error += 1
        print(f"Error in {file_name}: {repr(e)}")
        return

    print(f"Adding {file_name} to {combined_name}")

    with open(combined_name, "w") as f:
        f.write(json.dumps(combined, indent=2, ensure_ascii=False))


files = []
for arg in sys.argv[1:]:
    if arg.endswith(".json"):
        files.append(arg)
    else:
        files.extend(glob.glob(arg + "/*.json"))
print(len(files))

for idx, f in enumerate(files):
    if idx % 500 == 0:
        print(idx, stats.__dict__)
    process_file(f)

print(stats.__dict__)



#!/usr/bin/env python3

import sys
import json
import glob
import os
import random
import re
import copy
from jsonschema import Draft202012Validator, validate


def process_file(file_name):
    with open(file_name) as f:
        pos_data = json.loads(f.read())

    schema = pos_data["schema"]
    tests = pos_data["tests"]

    json_filename = file_name.replace("unique_tests", "json").replace("---", "/")

    with open(json_filename) as f:
        schema_2 = json.loads(f.read())

    if json.dumps(schema) != json.dumps(schema_2):
        with open(json_filename, "w") as f:
            f.write(json.dumps(schema, indent=4, ensure_ascii=False))


files = []
for arg in sys.argv[1:]:
    if arg.endswith(".json"):
        files.append(arg)
    else:
        files.extend(glob.glob(arg + "/*.json"))

for idx, f in enumerate(files):
    process_file(f)



#!/usr/bin/env python3

import sys
import json
import glob
import os
import random
import time
import resource

import xgrammar as xgr
import numpy as np
from transformers import AutoTokenizer, AutoConfig

output_path = "tmp/xgr/"


def time_us(prev: float) -> int:
    return int((time.monotonic() - prev) * 1000000)


def process_file(file: str):
    id = os.path.basename(file)
    output_name = output_path + id
    if os.path.exists(output_name):
        return None

    with open(output_name, "w") as f:
        f.write(json.dumps({ "pending_file": 1 }, indent=2))

    with open(file) as f:
        pos_data = json.loads(f.read())

    schema = json.dumps(pos_data["schema"])
    token_bitmask = xgr.allocate_token_bitmask(1, tokenizer_info.vocab_size)

    all_mask_us = []
    status = {
        "id": id,
        "ttfm_us": 0,
        "max_ttfm_us": 0,
        "masks_us": 0,
        "max_mask_us": 0,
        "num_tokens": 0,
        "num_tests": len(pos_data["tests"]),
        "all_mask_us": all_mask_us,
        "num_valid_tests": 0,
        "num_invalid_tests": 0,
    }

    try:
        t0 = time.monotonic()
        compiled_grammar = compiler.compile_json_schema(
            schema, any_whitespace=True, strict_mode=False
        )
        matcher = xgr.GrammarMatcher(compiled_grammar)
    except Exception as e:
        status["compile_error"] = repr(e)
        with open(output_name, "w") as f:
            f.write(json.dumps(status, indent=2))
        return status

    status["ttfm_us"] = time_us(t0)
    status["max_ttfm_us"] = status["ttfm_us"]

    masks_us = 0
    max_mask_us = 0
    num_tokens = 0

    for i, test in enumerate(pos_data["tests"]):
        instance = json.dumps(test["data"], indent=None)
        tokens = tokenizer.encode(instance, add_special_tokens=False)

        t1 = time.monotonic()
        accepted = True
        for tidx, t in enumerate(tokens):
            t2 = time.monotonic()
            matcher.fill_next_token_bitmask(token_bitmask)
            ok = matcher.accept_token(t)
            mask_time = time_us(t2)
            num_tokens += 1
            masks_us += mask_time
            all_mask_us.append(mask_time)
            if mask_time > max_mask_us:
                max_mask_us = mask_time
            if not ok:
                accepted = False
                break

        if accepted and not test["valid"]:
            status["validation_error"] = f"test #{i}: should reject but didn't"
        elif not accepted and test["valid"]:
            status["validation_error"] = f"test #{i}: should accept but didn't"
        else:
            if test["valid"]:
                status["num_valid_tests"] += 1
            else:
                status["num_invalid_tests"] += 1

    status["masks_us"] = masks_us
    status["max_mask_us"] = max_mask_us
    status["num_tokens"] = num_tokens

    with open(output_name, "w") as f:
        f.write(json.dumps(status, indent=2))
    return status

def main():
    global tokenizer_info, compiler, tokenizer

    limit_gb = 32
    limit_bytes = limit_gb * 1024 * 1024 * 1024
    resource.setrlimit(resource.RLIMIT_AS, (limit_bytes, limit_bytes))

    # Get tokenizer info
    model_id = "meta-llama/Llama-3.1-8B-Instruct"
    tokenizer = AutoTokenizer.from_pretrained(model_id)
    config = AutoConfig.from_pretrained(model_id)
    # This can be larger than tokenizer.vocab_size due to paddings
    full_vocab_size = config.vocab_size
    tokenizer_info = xgr.TokenizerInfo.from_huggingface(
        tokenizer, vocab_size=full_vocab_size
    )
    compiler = xgr.GrammarCompiler(tokenizer_info, max_threads=1)

    files = []
    for arg in sys.argv[1:]:
        if arg.endswith(".json"):
            files.append(arg)
        else:
            files.extend(glob.glob(arg + "/*.json"))
    print(len(files), file=sys.stderr)
    random.shuffle(files)

    os.makedirs(output_path, exist_ok=True)

    for f in files:
        print(f, file=sys.stderr)
        process_file(f)

main()


#!/usr/bin/env python3

import json
import sys

objs = []

with open(sys.argv[1]) as f:
    for l in f.readlines():
        if not l:
            continue
        j = json.loads(l)
        buf = ""
        for outline in j["output"].split("\n"):
            if outline.startswith("RESULT: "):
                parsed = json.loads(outline[len("RESULT: "):])
                parsed["warnings"] = buf
                objs.append(parsed)
                print(parsed.get("compile_time",0), parsed["file"])
                buf = ""
            else:
                buf += outline + "\n"

with open(sys.argv[2], "w") as f:
    f.write(json.dumps(objs, indent=4))


#!/usr/bin/env python3

import sys
import json
import glob
import os
import random
import re
from jsonschema import Draft202012Validator, validate

combined_base = os.environ.get("JSB_DATA") + "/unique_tests"
invalid_marker = "Please modify the instance to make it invalid according to the schema"


class Stats:
    def __init__(self):
        self.files = 0
        self.responses = 0
        self.duplicate_test = 0
        self.server_error = 0
        self.json_error = 0
        self.json_error_non_length = 0
        self.validation_error = 0
        self.invalidation_error = 0
        self.not_negative = 0
        self.tests_added = 0


stats = Stats()


def process_file(output_name):
    file_base = output_name.split("/")[-1]

    stats.files += 1

    combined_name = f"{combined_base}/{file_base}"

    if not os.path.exists(combined_name):
        print(f"File {combined_name} does not exist")
        return

    with open(output_name) as f:
        resp_file = json.loads(f.read())

    resps = [resp_file]
    if "responses" in resp_file:
        resps = resp_file["responses"]

    with open(combined_name) as f:
        try:
            combined = json.loads(f.read())
        except:
            print("Failed to load", combined_name)
            return

    existing_tests = set(json.dumps(t["data"]) for t in combined["tests"])

    schema = combined["schema"]
    # Draft202012Validator.check_schema(schema)

    num_added = 0

    for resp in resps:
        if resp.get("error", None):
            stats.server_error += 1
            continue

        rs = resp["choices"][0]["message"]["content"]
        try:
            r = json.loads(rs)
        except:
            stats.json_error += 1
            if resp["choices"][0]["finish_reason"] != "length":
                if resp["choices"][0].get("llg_logs", None):
                    print("non-length-llg", output_name, combined_name)
                else:
                    print("non-length", output_name, combined_name)
            continue

        if json.dumps(r) in existing_tests:
            stats.duplicate_test += 1
            continue

        test_is_valid = True
        prompt = resp.get("expanded_prompt", "")

        resp_meta = resp.get("meta", {})
        resp_meta_valid = resp_meta.get("valid", None)
        if resp_meta_valid is not None:
            test_is_valid = resp_meta_valid
        else:
            if invalid_marker in prompt:
                test_is_valid = False

        test = {
            "description": "llama 70b generated positive",
            "valid": test_is_valid,
            "data": r,
        }

        if not test_is_valid:
            description = "llama-70b generated negative"
            m = re.search(
                r"violate a constraint introduced by (.+?) in the schema", prompt
            )
            if m:
                description += "; focus on " + m.group(1)
            test["description"] = description

        try:
            validate(r, schema, format_checker=Draft202012Validator.FORMAT_CHECKER)
            if not test_is_valid:
                stats.invalidation_error += 1
                # print("invalidation error", output_name, combined_name)
                continue
        except Exception as e:
            if test_is_valid:
                stats.validation_error += 1
                # print("validation error", output_name, combined_name)
                continue
            else:
                test["python_error"] = str(e)

        combined["tests"].append(test)
        stats.tests_added += 1
        num_added += 1

    if num_added > 0:
        try:
            combined_bytes = json.dumps(combined, indent=2, ensure_ascii=False).encode("utf8")
        except Exception as e:
            # this may fail due to utf8 encoding issues
            print("Failed to dump", combined_name, e)
            return
        with open(combined_name, "bw") as f:
            f.write(combined_bytes)


files = []
for arg in sys.argv[1:]:
    if arg.endswith(".json"):
        files.append(arg)
    else:
        files.extend(glob.glob(arg + "/*.json"))
print(len(files))

for idx, f in enumerate(files):
    if idx % 500 == 0:
        print(idx, stats.__dict__)
    process_file(f)

print(stats.__dict__)



#!/usr/bin/env python3

from typing import Any, List
import sys
import json
import glob
import os
import random
import re
from jsonschema import Draft202012Validator, validate
import tokenizers
import llguidance
import copy


class PhiTokenizer:
    _ll_tokenizer = None
    _instance = None

    @staticmethod
    def instance():
        if PhiTokenizer._instance is None:
            PhiTokenizer._instance = PhiTokenizer()
        return PhiTokenizer._instance

    @staticmethod
    def ll_tokenizer():
        if PhiTokenizer._ll_tokenizer is None:
            PhiTokenizer._ll_tokenizer = llguidance.LLTokenizer(
                llguidance.TokenizerWrapper(PhiTokenizer.instance())
            )
        return PhiTokenizer._ll_tokenizer

    def tokenize_str(self, s: str) -> List[int]:
        return self.hf_tokenizer.encode(s).ids

    def __init__(self) -> None:
        self.hf_tokenizer: tokenizers.Tokenizer = tokenizers.Tokenizer.from_pretrained(
            "microsoft/Phi-3-mini-128k-instruct"
        )
        empty = self.tokenize_str("")
        if empty:
            self.bos_token_id = empty[0]
        else:
            self.bos_token_id = None
        eos = self.tokenize_str("</s>")
        assert len(eos) == 1
        self.eos_token_id = eos[0]
        self.tokens = []
        for i in range(self.hf_tokenizer.get_vocab_size()):
            t: str = self.hf_tokenizer.id_to_token(i)
            if t.startswith("<0x"):
                self.tokens.append(bytes([int(t[3:5], 16)]))
            else:
                t = t.replace("▁", " ")
                self.tokens.append(t.encode("utf-8"))
        assert len(self.tokens) == self.hf_tokenizer.get_vocab_size()

    def __call__(self, s):
        return self.tokenize_str(s)


class Stats:
    def __init__(self):
        self.files = 0
        self.grammar_error = 0
        self.valid_ok = 0
        self.valid_error = 0
        self.invalid_ok = 0
        self.invalid_error = 0
        self.num_exn = 0
        self.num_reorder_errors = 0
        self.num_half_wrong = 0


stats = Stats()
magic_key = "$$$magic_key_for_reordering$$$"


class TokensInvalid(Exception):
    pass


class ReorderError(Exception):
    pass


def remove_magic_key(obj):
    if isinstance(obj, dict):
        if magic_key in obj:
            del obj[magic_key]
        for k, v in obj.items():
            remove_magic_key(v)
    elif isinstance(obj, list):
        for v in obj:
            remove_magic_key(v)


# this assumes we don't force the canonical tokenization on commas, quotes etc.
# we could always just use byte tokenizer
def reorder_json(test: dict, interp: llguidance.LLInterpreter):
    interp = interp.deep_copy()
    text_buf = ""
    text_committed = ""
    dbg = False
    ll_tok = PhiTokenizer.ll_tokenizer()

    def save_state():
        return text_buf, text_committed, interp.deep_copy()

    def restore_state(state):
        nonlocal text_buf, text_committed, interp
        text_buf, text_committed, interp = state

    def stringify(v):
        return json.dumps(v, ensure_ascii=False, indent=None, separators=(",", ":"))

    def flush_tokens():
        nonlocal text_buf, text_committed
        if text_buf == "":
            return
        tokens = PhiTokenizer.ll_tokenizer().tokenize_str(text_buf)
        for idx, token in enumerate(tokens):
            n = interp.validate_tokens_raw([token])
            if n != 1:
                if dbg:
                    print(
                        "FAIL", text_committed, ll_tok.dbg_tokens(tokens[0 : idx + 1])
                    )
                raise TokensInvalid()
            interp.commit_token(token)
        text_committed += text_buf
        if dbg:
            print("OK", text_committed)
        text_buf = ""

    def append_text(text: str):
        nonlocal text_buf
        text_buf += text

    def validate_text(text: str):
        flush_tokens()
        tokens = PhiTokenizer.ll_tokenizer().tokenize_str(text)
        if interp.validate_tokens_raw(tokens) != len(tokens):
            if dbg:
                print("FAIL_V", text_committed, text)
            return False
        if dbg:
            print("OK_V", text_committed, text)
        return True

    def check(obj):
        if isinstance(obj, dict):
            if magic_key in obj:
                append_text(obj[magic_key])
            else:
                flush_tokens()
                committed_len = len(text_committed)
                append_text("{")
                key_order = []
                keys_left = list(obj.keys())
                if not keys_left:
                    append_text("}")
                    flush_tokens()
                while keys_left:
                    is_ok = False
                    possible_keys = [
                        k for k in keys_left if validate_text(stringify(k) + ":")
                    ]
                    for key in possible_keys:
                        prev_state = save_state()
                        try:
                            append_text(stringify(key) + ":")
                            check(obj[key])
                            # make sure we can still advance to all future keys
                            for key2 in possible_keys:
                                if key2 != key:
                                    if not validate_text("," + stringify(key2) + ":"):
                                        raise TokensInvalid()
                            if len(keys_left) == 1:
                                append_text("}")
                            else:
                                append_text(",")
                            flush_tokens()

                            # all good, we can continue
                            keys_left.remove(key)
                            key_order.append((key, obj[key]))
                            is_ok = True
                            break
                        except TokensInvalid:
                            restore_state(prev_state)
                            continue
                    if not is_ok:
                        raise ReorderError(
                            "no key can be advanced\n"
                            + stringify(obj)
                            + "\n\ntext_committed: "
                            + text_committed
                            + "\n\nkeys_left: "
                            + str(keys_left)
                            + "\npossible_keys: "
                            + str(possible_keys)
                            + "\nkey_order: "
                            + str(key_order)
                        )
                obj.clear()
                for k, v in key_order:
                    obj[k] = v
                if text_buf != "":
                    raise ValueError("text_buf not empty", text_buf)
                obj_str = text_committed[committed_len:]
                obj[magic_key] = obj_str
        elif isinstance(obj, list):
            append_text("[")
            for idx, v in enumerate(obj):
                if idx > 0:
                    append_text(",")
                check(v)
            append_text("]")
        else:
            append_text(stringify(obj))

    interp.start_without_prompt()

    root_obj = copy.deepcopy(test["data"])
    try:
        check(root_obj)
        flush_tokens()
        if not interp.is_accepting():
            raise ValueError("reordering failed", stringify(root_obj))
        remove_magic_key(root_obj)
        test["data"] = root_obj
        # print("reordering succeeded", stringify(root_obj))
    except ReorderError as e:
        if not test["valid"]:
            # kind of expected, save what we can
            remove_magic_key(root_obj)
            test["data"] = root_obj
            stats.num_half_wrong += 1
            return
        raise e
    except TokensInvalid:
        if not test["valid"]:
            # kind of expected, save what we can
            remove_magic_key(root_obj)
            test["data"] = root_obj
            stats.num_half_wrong += 1
            return
        raise ValueError("reordering failed", stringify(root_obj))


def remove_constraints(obj):
    if isinstance(obj, dict):
        if "properties" in obj:
            for k, v in obj["properties"].items():
                remove_constraints(v)
            return
        for k in [
            "minimum",
            "maximum",
            "exclusiveMinimum",
            "exclusiveMaximum",
            "multipleOf",
            "pattern",
            "minLength",
            "maxLength",
            "format",
        ]:
            if k in obj:
                del obj[k]
        for k, v in obj.items():
            remove_constraints(v)
    elif isinstance(obj, list):
        for v in obj:
            remove_constraints(v)


def mk_interp(schema):
    return llguidance.LLInterpreter(
        PhiTokenizer.ll_tokenizer(),
        json.dumps(
            {
                "grammars": [
                    {
                        "json_schema": schema,
                    }
                ]
            },
            ensure_ascii=False,
        ),
        log_level=0,
        enable_ff_tokens=False,
        enable_backtrack=False,
    )


def process_file(file_name):
    file_base = file_name.split("/")[-1]

    stats.files += 1

    with open(file_name) as f:
        pos_data = json.loads(f.read())

    schema = pos_data["schema"]
    tests = pos_data["tests"]

    schema_simplified = copy.deepcopy(schema)
    remove_constraints(schema_simplified)

    try:
        interp0 = mk_interp(schema)
    except Exception as e:
        # print("interpreter creation error", file_name, str(e))
        stats.grammar_error += 1
        return

    interp1 = mk_interp(schema_simplified)

    num_reordered = 0

    for idx, test in enumerate(tests):
        test_str = json.dumps(test["data"], ensure_ascii=False, indent=None)
        ll_tok = PhiTokenizer.ll_tokenizer()
        tokens = ll_tok.tokenize_str(test_str)
        # print(ll_tok.dbg_tokens(tokens))

        interp = interp0.deep_copy()

        interp.start_without_prompt()

        try:
            tests_valid = True
            for tidx, token in enumerate(tokens):
                n = interp.validate_tokens_raw([token])
                if n == 0:
                    tests_valid = False
                    if test["valid"]:
                        # print(
                        #     "test fails",
                        #     file_name,
                        #     idx,
                        #     ll_tok.dbg_tokens(tokens[0 : tidx + 1]),
                        # )
                        reorder_json(test, interp0)
                        stats.valid_error += 1
                        num_reordered += 1
                    else:
                        # first try with a simplified schema, that
                        # doesn't have string/number constraints
                        reorder_json(test, interp1)
                        reorder_json(test, interp0)
                        num_reordered += 1

                    break
                interp.commit_token(token)

            if tests_valid and not interp.is_accepting():
                tests_valid = False
                if test["valid"]:
                    print("test fails at the end", file_name, idx)
                    stats.valid_error += 1

            if tests_valid:
                if test["valid"]:
                    stats.valid_ok += 1
                else:
                    print("invalid test shouldn't pass", file_name, idx)
                    stats.invalid_error += 1
            else:
                if not test["valid"]:
                    stats.invalid_ok += 1
        except ReorderError as e:
            print("reorder error", file_name, idx, str(e))
            stats.num_reorder_errors += 1
        except Exception as e:
            stats.num_exn += 1
            # raise e
            if "consider making your grammar left-recursive" not in str(e):
                print("validation error", file_name, idx, repr(e))

    if num_reordered > 0:
        with open(file_name, "w") as f:
            f.write(json.dumps(pos_data, ensure_ascii=False, indent=2))


files = []
for arg in sys.argv[1:]:
    if arg.endswith(".json"):
        files.append(arg)
    else:
        files.extend(glob.glob(arg + "/*.json"))
print(len(files))
files.sort()

for idx, f in enumerate(files):
    if idx > 0 and idx % 500 == 0:
        print(idx, stats.__dict__, file=sys.stderr)
    process_file(f)

print(stats.__dict__)



#!/usr/bin/env python3

import sys
import json
import glob
import os
import random
import re
from jsonschema import Draft202012Validator, validate

class Stats:
    def __init__(self):
        self.files = 0
        self.responses = 0
        self.server_error = 0
        self.json_error = 0
        self.json_error_non_length = 0
        self.validation_error = 0
        self.invalidation_error = 0
        self.not_negative = 0
        self.negative_added = 0


stats = Stats()


def process_file(file_name):
    file_base = file_name.split("/")[-1]

    stats.files += 1

    with open(file_name) as f:
        pos_data = json.loads(f.read())
   
    schema = pos_data["schema"]
    tests = pos_data["tests"]

    Draft202012Validator.check_schema(schema)

    for idx, test in enumerate( tests ):
        try:
            validate(test["data"], schema, format_checker=Draft202012Validator.FORMAT_CHECKER)
            if not test["valid"]:
                print("positive already there", file_name, idx)
                stats.invalidation_error += 1
        except Exception as e:
            if test["valid"]:
                stats.validation_error += 1
                print("validation error", file_name, idx, str(e))
            else:
                test["python_error"] = str(e)

    with open(file_name, "w") as f:
        f.write(json.dumps(pos_data, indent=2, ensure_ascii=False))


files = []
for arg in sys.argv[1:]:
    if arg.endswith(".json"):
        files.append(arg)
    else:
        files.extend(glob.glob(arg + "/*.json"))
print(len(files))

for idx, f in enumerate( files ):
    if idx % 500 == 0:
        print(idx, stats.__dict__)
    process_file(f)

print(stats.__dict__)



#!/usr/bin/env python3

import sys
import json
import glob
import os
import random
import time
import concurrent.futures
import subprocess
from typing import List

positive_base = os.environ.get("HOME") + "/src/json-data/positive"
output_base = os.environ.get("HOME") + "/src/json-data/xgr_output"

def process_file(files: List[str]):
    timeout=60
    try:
        command = ["python", "xgr_test.py"] + files
        
        result = subprocess.run(
            command,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            encoding="utf-8",
            timeout=timeout
        )
        
        # Collect combined output
        output = result.stdout
        
        return {
            "output": output,
            "return_code": result.returncode
        }
    except subprocess.TimeoutExpired as e:
        output = e.output or ""
        # decode output if needed
        if isinstance(output, bytes):
            output = output.decode("utf-8", errors="replace")
        return {
            "output": output,
            "error": "TimeoutExpired",
            "message": f"The process exceeded the {timeout}-second timeout."
        }
    except Exception as e:
        return {
            "error": "ExecutionError",
            "message": str(e)
        }

file_name = []
for arg in sys.argv[1:]:
    if arg.endswith(".json"):
        file_name.append(arg)
    else:
        file_name.extend(glob.glob(arg + "/*.json"))
print(len(file_name), file=sys.stderr)
missing_files = []
for f in file_name:
    file_base = f.split("/")[-1]
    output_name = f"{output_base}/{file_base}"
    if not os.path.exists(output_name):
        missing_files.append(f)
print(len(missing_files), file=sys.stderr)
random.shuffle(missing_files)

if len(missing_files) < 10:
    print(missing_files)
    sys.exit(0)

chunk_size = 10
chunks = []
for i in range(0, len(missing_files), chunk_size):
    chunks.append(missing_files[i:i + chunk_size])

log_file = f"{output_base}/log.txt"
cnt = 0

with concurrent.futures.ThreadPoolExecutor(max_workers=40) as executor:
    futures = {executor.submit(process_file, f): f for f in chunks}
    for future in concurrent.futures.as_completed(futures):
        file_name = futures[future]
        try:
            r = future.result()
            print(file_name)
            # print(f"OK: {files}")
        except Exception as e:
            print(f"ERROR: {file_name}", repr(e))



#!/usr/bin/env python3

import requests
import sys
import json
import glob
import os
import random
import concurrent.futures


output_base = os.environ.get("JSB_DATA") + "/work/responses"

url = "http://localhost:3001/v1/chat/completions"

gen_negative = False
gen_constrained = False


def feature_name(feature):
    if feature == "type:[]":
        return "type keyword with an array of types"
    if feature == "additionalProperties:object":
        return "additionalProperties keyword with an object schema"
    return feature + " keyword"


singleton_features = [
    "type:[]",
    "pattern",
    "oneOf",
    "format",
    "allOf",
    "anyOf",
    "patternProperties",
    "dependencies",
    "additionalItems",
    "additionalProperties:object",
    "multipleOf",
    "uniqueItems",
    "propertyNames",
    "contains",
]

group_features = [
    ["minProperties", "maxProperties"],
    ["minimum", "maximum", "exclusiveMinimum", "exclusiveMaximum"],
    ["minLength", "maxLength"],
    ["minItems", "maxItems"],
    ["if", "then", "else"],
]


def gen_example(schema: dict, instance: dict | None, info: str, constrained: bool):
    if instance is None:
        valid = True
        prompt = (
            "Please generate an example data that validates against the following schema:\n"
            + json.dumps(schema, indent=4, ensure_ascii=False)
        )
    else:
        valid = False
        constrained = False
        prompt = (
            ""
            + f"You will be given a JSON schema an example instance of it. "
            + f"You will then modify the instance a little bit to violate the schema.\n"
            + "\n\nHere is the schema:\n"
            + json.dumps(schema, indent=4, ensure_ascii=False)
            + "\n\nHere is the valid instance:\n"
            + json.dumps(instance, indent=4, ensure_ascii=False)
            + "\n\nPlease modify the instance to make it invalid according to the schema. Focus on corner cases.\n"
            + info
        )
    if len(prompt) > 200_000:
        return {"error": f"Prompt too long, {len(prompt)}"}

    # empty schema to allow any JSON response
    format = {"schema": {}}
    if constrained:
        format = {
            "strict": True,
            "schema": schema,
        }

    req = {
        "model": "model",
        "messages": [
            {
                "role": "system",
                "content": "You are an expert in generating testcases for JSON schema validators.",
            },
            {"role": "user", "content": prompt},
        ],
        "return_expanded_prompt": True,
        "include_json_schema_in_prompt": False,
        # "llg_log_level": "verbose",
        "response_format": {
            "type": "json_schema",
            "json_schema": format,
        },
        "max_tokens": 8000,
        "temperature": 0.7,
    }

    headers = {
        "Content-Type": "application/json",
    }

    response = requests.post(url, headers=headers, json=req)
    # response.raise_for_status()
    r = response.json()
    r["meta"] = {
        "valid": valid,
        "info": info,
        "prompt": prompt,
    }
    return r


def process_file(file: str):
    file_base = file.split("/")[-1]
    output_name = f"{output_base}/{file_base}"
    if os.path.exists(output_name):
        return

    with open(file) as f:
        test_file = json.loads(f.read())
    last_valid_test = None

    if gen_negative:
        for t in test_file["tests"]:
            if t["valid"]:
                last_valid_test = t
        if last_valid_test is None:
            return

    stats = test_file["meta"]
    feature_names = set()
    for k in stats["features"]:
        if k in singleton_features:
            feature_names.add(feature_name(k))
        else:
            for lst in group_features:
                if k in lst:
                    lst2 = [e for e in lst if e in stats["features"]]
                    name = " or ".join([feature_name(e) for e in lst2])
                    feature_names.add(name)
                    break

    feature_list = [
        f"Try to modify the instance to detect bugs in a validator that doesn't support {f}. In other words, try to violate a constraint introduced by {f} in the schema.\n"
        for f in feature_names
    ]
    feature_list.append("")
    responses = []

    if gen_negative:
        for f in feature_list:
            r = gen_example(
                schema=test_file["schema"],
                instance=last_valid_test["data"],
                info=f,
                constrained=False,
            )
            responses.append(r)
    else:
        # right now, not using features when generating positive examples
        r = gen_example(
            schema=test_file["schema"],
            instance=None,
            info="",
            constrained=gen_constrained,
        )
        responses.append(r)

    res = {
        "pos_data": test_file,
        "responses": responses,
    }
    output = json.dumps(res, indent=4, ensure_ascii=False)
    with open(output_name, "w") as f:
        f.write(output)
    print(f"Wrote {output_name}  {len(output)} bytes")


args = sys.argv[1:]
while len(args) > 0 and args[0].startswith("-"):
    arg = args.pop(0)
    if arg == "-n":
        gen_negative = True
    elif arg == "-c":
        gen_constrained = True
    else:
        raise ValueError(f"Unknown option {arg}")

files = []
for arg in args:
    if arg.endswith(".json"):
        files.append(arg)
    else:
        files.extend(glob.glob(arg + "/*.json"))
print(
    len(files),
    "files to process",
    "negative" if gen_negative else "positive",
    "constrained" if gen_constrained else "unconstrained",
)
random.shuffle(files)

try:
    os.makedirs(output_base)
except FileExistsError:
    pass

with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
    futures = {executor.submit(process_file, f): f for f in files}
    for future in concurrent.futures.as_completed(futures):
        file_name = futures[future]
        try:
            future.result()  # This will raise exceptions if any occurred
        except Exception as e:
            print(f"Error processing {file_name}: {repr(e)}")



#!/usr/bin/env python3

import json
import math
import glob

output_path = "tmp/xgr/"
llg = False


class Stats:
    def __init__(self) -> None:
        self.ttfm_us = 0
        self.max_ttfm_us = 0
        self.masks_us = 0
        self.masks_us_under_10ms = 0
        self.num_masks_under_10ms = 0
        self.avg_masks_under_10ms = 0
        self.masks_us_over_10ms = 0
        self.num_masks_over_10ms = 0
        self.avg_masks_over_10ms = 0
        self.max_mask_us = 0
        self.num_tokens = 0
        self.num_schemas = 0
        self.num_schemas_ok = 0
        self.num_compilation_errors = 0
        self.num_validation_errors = 0
        self.num_tests = 0
        self.num_valid_tests = 0
        self.num_invalid_tests = 0


def log_fraction_plot(times: list[int]):
    times.sort()
    cutoff = 1
    mult = 1.3
    count = 0
    csv = "cutoff time,frac left\n"
    total = len(times)
    for t in times:
        while t > cutoff:
            csv += f"{cutoff/1000.0},{(total - count)/total}\n"
            cutoff = int(cutoff * mult) + 1
        count += 1
    return csv


def histogram_position(us: int):
    return int(math.floor(math.log10(max(1, us - 1))))

def us_to_str(us: int):
    if us < 1000:
        return f"{us}us"
    if us < 1000000:
        return f"{us//1000}ms"
    return f"{us//1000000}s"


def main():
    if llg:
        files = ["tmp/llg_results.json"]
    else:
        files = glob.glob(output_path + "*.json")
        files = sorted(files)
    stats = Stats()
    ttfm_us = []
    all_masks_us = []
    histogram_us = [0] * 10
    histogram_num = [0] * 10
    for f in files:
        with open(f) as f:
            data = json.load(f)
        elts = [data]
        if isinstance(data, list):
            elts = data
        for data in elts:
            if "num_tests" not in data:
                continue
            stats.num_schemas += 1
            stats.num_tests += data["num_tests"]
            if "compile_error" in data:
                stats.num_compilation_errors += 1
            else:
                stats.ttfm_us += data["ttfm_us"]
                ttfm_us.append(data["ttfm_us"])
                stats.max_ttfm_us = max(data["max_ttfm_us"], stats.max_ttfm_us)
                if "masks_us" in data:
                    stats.masks_us += data["masks_us"]
                    stats.max_mask_us = max(data["max_mask_us"], stats.max_mask_us)
                    stats.num_tokens += data["num_tokens"]
                if "validation_error" in data:
                    stats.num_validation_errors += 1
                else:
                    stats.num_schemas_ok += 1
                stats.num_valid_tests += data["num_valid_tests"]
                stats.num_invalid_tests += data["num_invalid_tests"]
                all_masks_us.extend(data["all_mask_us"])
                for us in data["all_mask_us"]:
                    p = histogram_position(us)
                    histogram_us[p] += us
                    histogram_num[p] += 1
                    if us < 10000:
                        stats.masks_us_under_10ms += us
                        stats.num_masks_under_10ms += 1
                    else:
                        stats.masks_us_over_10ms += us
                        stats.num_masks_over_10ms += 1
    stats.avg_masks_under_10ms = stats.masks_us_under_10ms // stats.num_masks_under_10ms
    stats.avg_masks_over_10ms = stats.masks_us_over_10ms // stats.num_masks_over_10ms
    print(json.dumps(stats.__dict__, indent=2))
    with open("tmp/xgr_ttfm_us.csv", "w") as f:
        f.write(log_fraction_plot(ttfm_us))
    with open("tmp/xgr_masks_us.csv", "w") as f:
        f.write(log_fraction_plot(all_masks_us))
    
    all_masks_us.sort()
    for p in [25, 50, 75, 90, 95, 99, 99.9, 99.99]:
        print(f"{'p' + str(p):7}, {all_masks_us[int(len(all_masks_us) * p / 100)]}")

    num_masks = sum(histogram_num)
    h_csv = "above us,frac\n"
    for i in range(10)[1:]:
        frac = sum(histogram_num[i:]) * 100 / num_masks
        h_csv += f"{us_to_str(10**i):10}"
        h_csv += f","
        h_csv += f"{frac:1.15}"
        h_csv += f"\n"
    with open("tmp/xgr_histogram.csv", "w") as f:
        f.write(h_csv)
    print(h_csv)


main()
